
import sys
import os
import threading
from contextlib import contextmanager, asynccontextmanager
from typing import Callable, Optional

class CaptureOutput:
    def __init__(self, callback: Callable[[str], None]):
        self.callback = callback
        self._stdout = sys.stdout
        self._stderr = sys.stderr
        self._stdout_pipe_read = None
        self._stdout_pipe_write = None
        self._stderr_pipe_read = None
        self._stderr_pipe_write = None
        self._stdout_thread = None
        self._stderr_thread = None
        self._stop_event = threading.Event()

    def _redirect_output(self, pipe_read, prefix):
        while not self._stop_event.is_set():
            line = pipe_read.readline()
            if line:
                self.callback(f"[{prefix}] {line.strip()}")

    def _start_threads(self):
        self._stdout_pipe_read, self._stdout_pipe_write = os.pipe()
        self._stderr_pipe_read, self._stderr_pipe_write = os.pipe()

        sys.stdout = os.fdopen(self._stdout_pipe_write, "w")
        sys.stderr = os.fdopen(self._stderr_pipe_write, "w")

        stdout_read = os.fdopen(self._stdout_pipe_read, "r")
        stderr_read = os.fdopen(self._stderr_pipe_read, "r")

        self._stdout_thread = threading.Thread(target=self._redirect_output, args=(stdout_read, "stdout"), daemon=True)
        self._stderr_thread = threading.Thread(target=self._redirect_output, args=(stderr_read, "stderr"), daemon=True)

        self._stdout_thread.start()
        self._stderr_thread.start()

    def _stop_threads(self):
        self._stop_event.set()
        sys.stdout.close()
        sys.stderr.close()
        sys.stdout = self._stdout
        sys.stderr = self._stderr

    @contextmanager
    def context(self):
        try:
            self._start_threads()
            yield
        finally:
            self._stop_threads()

    @asynccontextmanager
    async def async_context(self):
        try:
            self._start_threads()
            yield
        finally:
            self._stop_threads()

    def __enter__(self):
        self._start_threads()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._stop_threads()

    async def __aenter__(self):
        self._start_threads()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._stop_threads()
