from typing import Any
from nicegui import ui
import inspect
import json


class WidgetRegistry:
    def __init__(self):
        self._registry = {}

    def register(self, type_, widget_class):
        """Registers a widget class for a given type."""
        if type_ in self._registry:
            raise ValueError(f"Widget already registered for type {type_}")
        self._registry[type_] = widget_class

    def get_widget(self, type_):
        """Retrieves the widget class for a given type or falls back to Any."""
        # Direct match
        widget_class = self._registry.get(type_)
        if widget_class is not None:
            return widget_class
        # Fallback to Any
        return self._registry.get(Any)

    def render(self, type_, **kwargs):
        """Renders a widget for a given type."""
        widget_class = self.get_widget(type_)
        return widget_class(self, **kwargs)

    def render_dataclass(self, dataclass):
        """Renders a widget for a given dataclass."""
        container = ui.element()
        with container:
            for field in dataclass.__dataclass_fields__.values():
                default_value = getattr(dataclass, field.name)
                self.render(field.type, label=field.name, value=default_value)
        return container

    def render_function(self, function):
        """Renders a widget for a given function."""
        container = ui.element()
        params = {}
        with container:
            for parameter in inspect.signature(function).parameters.values():
                if parameter.default != inspect._empty:
                    params[parameter.name] = parameter.default
                self.render(parameter.annotation, label=parameter.name).bind_value(params, parameter.name)
            def run(event):
                ui.label(json.dumps(params))
            ui.button("Run", on_click=run)
        return container

    def __call__(self, type_):
        """Decorator for registering widgets."""

        def decorator(widget_class):
            self.register(type_, widget_class)
            return widget_class

        return decorator

    def __repr__(self):
        return f"WidgetRegistry({self._registry})"


widget_registry = WidgetRegistry()


@widget_registry(int)
class WidgetInt(ui.number):
    def __init__(self, registry, format="%d", precision=0, **kwargs):
        super().__init__(format=format, precision=precision, **kwargs)


@widget_registry(float)
class WidgetFloat(ui.number):
    def __init__(self, registry, format="%.2f", precision=2, **kwargs):
        super().__init__(format=format, precision=precision, **kwargs)


@widget_registry(str)
class WidgetStr(ui.input):
    def __init__(self, registry, **kwargs):
        super().__init__(**kwargs)


@widget_registry(Any)
class WidgetAny(ui.codemirror):
    def __init__(
        self,
        registry,
        label="",
        value="",
        language="json",
        theme="githubDark",
        **kwargs,
    ):
        value = json.dumps(value)
        ui.label(label)
        super().__init__(value=value, theme=theme, language=language, **kwargs)

    def set_value(self, value):
        self.value = json.dumps(value)

    def get_value(self):
        return json.loads(self.value)

